using Kafka.Public;
using Kafka.Public.Loggers;
using Newtonsoft.Json;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Reflection;
using System.Threading;
using System.Threading.Tasks;

namespace franz
{
    public class Telefranz
    {
        private static Telefranz instance = null;
        private static readonly object createLock = new object();

        static class _TelefranzConsumers<T> where T : gray_messages.message
        {
            internal static readonly Dictionary<Action<T>, Action<KafkaRecord<string, string>>> wrappings
                = new Dictionary<Action<T>, Action<KafkaRecord<string, string>>>();
        }
        private string handling_group { get; set; } = "Liszt";
        private gray_messages.global.report howToReport { get; set; }
        private ClusterClient clusterClient;
        private Dictionary<string, KafkaConsumer<string, string>> topicConsumers { get; set; }
             = new Dictionary<string, KafkaConsumer<string, string>>();
        private Dictionary<string, int> topicSubscribers { get; set; } = new Dictionary<string, int>();
        private Kafka.Public.Configuration clusterClientConfig = null;
        private static quieterLogger consoleLogger = new quieterLogger();
        private Telefranz(string name, string bootstrap_servers,
            List<string> commands = null, List<string> checks = null,
            List<string> errors = null, List<string> warnings = null)
        {
            var serializer = new StringSerializer();
            var deserializer = new StringDeserializer();
            var serializationConfig = new SerializationConfig() { SerializeOnProduce = true };
            serializationConfig.SetDefaultSerializers(serializer, serializer);
            serializationConfig.SetDefaultDeserializers(deserializer, deserializer);
            this.clusterClientConfig = new Kafka.Public.Configuration
            {
                Seeds = bootstrap_servers,
                ClientId = name,
                SerializationConfig = serializationConfig
            };
            this.clusterClient = new ClusterClient(clusterClientConfig, consoleLogger);

            handling_group = name;
            this.howToReport = new gray_messages.global.report()
            {
                name = name,
                host = Dns.GetHostName(),
                errors = errors ?? new List<string>(),
                warnings = warnings ?? new List<string>()
            };
            if (commands?.Count > 0)
            {
                howToReport.capabilites.commands = commands;
            }
            if (checks?.Count > 0)
            {
                howToReport.capabilites.checks = checks;
            }

            addHandler<gray_messages.global.sound_off>((m) =>
            {
                ProduceMessage(this.howToReport);
            });
            addHandler<gray_messages.global.stop>((m) =>
            {
                if (m.name == handling_group)
                {
                    Environment.Exit(0);
                }
            });
        }
        public static void Configure(string name, string bootstrap_servers,
            List<string> commands = null, List<string> checks = null,
            List<string> errors = null, List<string> warnings = null)
        {
            lock (createLock)
            {
                if (instance == null)
                {
                    instance = new Telefranz(name, bootstrap_servers, commands, checks, errors, warnings);
                }
            }

        }

        public static Telefranz Instance
        {
            get
            {
                lock (createLock)
                {
                    if (instance == null)
                    {
                        throw new NotInitializedException("Configure me first");
                    }
                }
                return instance;
            }
        }
        public void addHandler<T>(Action<T> theAction) where T : gray_messages.message
        {
            var topic = typeof(T).ToString();

            Action<KafkaRecord<string, string>> wrapped = (kr) =>
            {
                var deserialized = JsonConvert.DeserializeObject<T>(kr.Value);
                if (deserialized != null && deserialized.type == deserialized.GetType().ToString())
                {
                    try
                    {
                        theAction(deserialized);
                    }
                    catch (Exception e)
                    {
                        consoleLogger.LogError("if I don't catch this error the whole thing dies so...");
                        consoleLogger.LogError(JsonConvert.SerializeObject(e));
                    }
                }
            };
            if (!topicConsumers.ContainsKey(typeof(T).ToString()))
            {
                topicConsumers[topic] = new KafkaConsumer<string, string>(topic, clusterClient);
                topicConsumers[topic].ConsumeFromLatest();
                topicSubscribers.Add(topic, 0);
            }
            topicConsumers[topic].MessageReceived += wrapped;
            _TelefranzConsumers<T>.wrappings[theAction] = wrapped;
            topicSubscribers[topic]++;
        }
        public bool removeHandler<T>(Action<T> theAction) where T : gray_messages.message
        {
            topicConsumers[typeof(T).ToString()].MessageReceived -= _TelefranzConsumers<T>.wrappings[theAction];
            if (_TelefranzConsumers<T>.wrappings.ContainsKey(theAction))
            {
                _TelefranzConsumers<T>.wrappings.Remove(theAction);
                topicSubscribers[typeof(T).ToString()]--;
                if (topicSubscribers[typeof(T).ToString()] == 0)
                {
                    topicConsumers[typeof(T).ToString()].StopConsume();
                }
                return true;
            }
            else
            {
                return false;
            }
        }

        public void ProduceMessage<T>(T message) where T : gray_messages.message
        {
            Console.WriteLine(message.ToString());
            clusterClient.Produce(typeof(T).ToString(), message.ToString());
        }
    }
    public class NotInitializedException : Exception
    {
        public NotInitializedException(string message) : base(message) { }
    }
}